#include <sys/types.h>
#include <utime.h>
#include <time.h>
#include <stdio.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#include <netax25/ax25.h>
#include <netax25/axlib.h>
#include <netax25/axconfig.h>
#include <iostream>
using namespace std;



#define	MAX_PACKETLEN	512
#define	MAX_BUFLEN	2*MAX_PACKETLEN


static int backoff = -1;
static int ax25mode = -1;
static int debug = FALSE;
static int af_mode = AF_AX25;
static int window = 0;
static char *port = NULL;
int paclen = 0;
int fd;


int connect_to(char *address[])
{
  int fd = 0;
  int addrlen = 0;
  union {
    struct full_sockaddr_ax25 ax25;
  } sockaddr;

  char *digi;
  int one = debug;

  if (address[0] == NULL)
  {
    cerr << "call: too few arguments for AX.25" << endl;
    return ( -1);
  }
  if ((fd = socket(AF_AX25, SOCK_SEQPACKET, 0)) < 0)
  {
    cerr << "socket" << strerror(errno) << endl;
    return ( -1);
  }
  ax25_aton(ax25_config_get_addr(port), &sockaddr.ax25);
  if (sockaddr.ax25.fsa_ax25.sax25_ndigis == 0)
  {
    ax25_aton_entry(ax25_config_get_addr(port),
                    sockaddr.ax25.fsa_digipeater[0].
                    ax25_call);
    sockaddr.ax25.fsa_ax25.sax25_ndigis = 1;
  }
  sockaddr.ax25.fsa_ax25.sax25_family = AF_AX25;
  addrlen = sizeof(struct full_sockaddr_ax25);

  if (setsockopt
      (fd, SOL_AX25, AX25_WINDOW, &window,
       sizeof(window)) == -1)
  {
    cerr << "AX25_WINDOW" << strerror(errno) << endl;
    close(fd);
    return ( -1);
  }
  if (setsockopt
      (fd, SOL_AX25, AX25_PACLEN, &paclen,
       sizeof(paclen)) == -1)
  {
    cerr << "AX25_PACLEN" << strerror(errno) << endl;
    close(fd);
    return ( -1);
  }
  if (backoff != -1)
  {
    if (setsockopt
        (fd, SOL_AX25, AX25_BACKOFF, &backoff,
         sizeof(backoff)) == -1)
    {
      cerr << "AX25_BACKOFF" << strerror(errno) << endl;
      close(fd);
      return ( -1);
    }
  }
  if (ax25mode != -1)
  {
    if (setsockopt
        (fd, SOL_AX25, AX25_EXTSEQ, &ax25mode,
         sizeof(ax25mode)) == -1)
    {
      cerr << "AX25_EXTSEQ" << strerror(errno) << endl;
      close(fd);
      return ( -1);
    }
  }

  if (debug
      && setsockopt(fd, SOL_SOCKET, SO_DEBUG, &one,
                    sizeof(one)) == -1)
  {
    cerr << "SO_DEBUG" << strerror(errno) << endl;
    close(fd);
    return ( -1);
  }
  if (bind(fd, (struct sockaddr *) &sockaddr, addrlen) == -1)
  {
    cerr << "bind" << strerror(errno) << endl;
    close(fd);
    return ( -1);
  }

  if (ax25_aton_arglist
      ((const char **) address, &sockaddr.ax25) == -1)
  {
    close(fd);
    return ( -1);
  }
  sockaddr.ax25.fsa_ax25.sax25_family = AF_AX25;
  addrlen = sizeof(struct full_sockaddr_ax25);

  cerr << "Trying...\r" << flush;

  if (connect(fd, (struct sockaddr *) &sockaddr, addrlen))
  {
    cerr << endl << "connect" << strerror(errno) << endl;
    close(fd);
    return ( -1);
  }
  cerr << "*** Connected to " << address[0] << endl;
  ;
  return (fd);
}


int cmd_call(char *call[])
{
  if ((fd = connect_to(call)) == -1)
    return false;

  fcntl(fd, F_SETFL, O_NONBLOCK);
  fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);

  char writebuf[MAX_BUFLEN];
  long writebufstart = 0;
  long writebufend = 0;

  bool stdinclosed = false;

  while (true)
  {
    fd_set sock_read;
    FD_ZERO(&sock_read);
    if (!stdinclosed)
      if (writebufstart == writebufend)
        FD_SET(STDIN_FILENO, &sock_read);
    FD_SET(fd, &sock_read);

    fd_set sock_write;
    FD_ZERO(&sock_write);
    if (writebufstart != writebufend)
      FD_SET(fd, &sock_write);
    if ( -1 == select(fd + 1, &sock_read, &sock_write, NULL, NULL))
    {
      if (errno == EAGAIN)
        continue;

      cerr << "select" << strerror(errno) << endl;
      break;
    }


    if (FD_ISSET(fd, &sock_read))
    {
      char buf[MAX_BUFLEN];
      int bytes = read(fd, buf, MAX_BUFLEN);
      if (bytes == -1
          && errno != EWOULDBLOCK
          && errno != EAGAIN)
      {
        if (errno != ENOTCONN)
          cerr << "read" << strerror(errno) << endl;
        break;
      }
      for (int i = 0;i < bytes;i++)
        if (buf[i] == '\r')
          buf[i] = '\n';
      write(STDOUT_FILENO, buf, bytes);
    }

    if (FD_ISSET(fd, &sock_write))
    {
      long writtenbytes = write(fd, &writebuf[writebufstart], writebufend - writebufstart);
      if (writtenbytes == -1)
      {
        cerr << "write" << strerror(errno) << endl;
        break;
      }
      writebufstart += writtenbytes;
    }


    if (FD_ISSET(STDIN_FILENO, &sock_read))
    {
      writebufstart = 0;
      writebufend = read(STDIN_FILENO, writebuf, MAX_BUFLEN);
      if (writebufend == -1)
      {
        cerr << "read" << strerror(errno) << endl;
        break;
      }
      if (writebufend == 0)
      {
        cerr << "nothing to read, stdin closed?" << endl;
        stdinclosed = true;
      }

      for (int i = writebufstart;i < writebufend;i++)
        if (writebuf[i] == '\n')
          writebuf[i] = '\r';
    }



  }
  close(fd);
  fcntl(STDIN_FILENO, F_SETFL, 0);

  cerr << "*** Cleared" << endl;
  return false;
}


int main(int argc, char **argv)
{
  if (2>argc)
  {
    cerr << "usage: simplecall port callsign [[via] digipeaters...]" << endl;
    return 1;
  }
  port = argv[1];


  if (ax25_config_load_ports() == 0)
  {
    cerr << "call: no AX.25 port data configured" << endl;
    return 1;
  }
  if (ax25_config_get_addr(port) == NULL)
  {
    cerr << "bad port" << endl;
  }

  window = ax25_config_get_window(port);
  paclen = ax25_config_get_paclen(port);

  while (cmd_call(argv + optind + 1))
  {
    cerr << "Wait 60 sec before reconnect" << endl;
    sleep(60);
  }

  return 0;
}
